import torch
import torch.nn as nn

from copy import deepcopy
import time
import random

from measures import _reparam
from models import CustomResNet

from torch.utils.data import DataLoader
from dataset_helpers import default_collate


def sign(x):
    return (1, -1)[x < 0]


def stiffness(model, train_x, train_y, valid_x, valid_y, criterion, device):
    cos_values = list()

    train_model = _reparam(model)
    valid_model = deepcopy(train_model)

    train_x, train_y, valid_x, valid_y = train_x.to(device), train_y.to(device), valid_x.to(device), valid_y.to(device)
    train_model.to(device)
    valid_model.to(device)

    train_model.eval()
    valid_model.eval()

    train_loss = criterion(train_model(train_x), train_y)
    valid_loss = criterion(valid_model(valid_x), valid_y)

    for i in range(train_loss.size(0)):
        dot_product = torch.tensor(0.).to(device)
        train_product = torch.tensor(0.).to(device)
        valid_product = torch.tensor(0.).to(device)

        train_model.zero_grad()
        valid_model.zero_grad()

        train_loss[i].backward(retain_graph=True)
        valid_loss[i].backward(retain_graph=True)

        train_params = list(train_model.named_parameters())
        valid_params = list(valid_model.named_parameters())

        train_params.sort(key=lambda element: element[0])
        valid_params.sort(key=lambda element: element[0])

        for j in range(len(train_params)):
            grad1 = train_params[j][1].grad.flatten()
            grad2 = valid_params[j][1].grad.flatten()
            dot_product += torch.dot(grad1, grad2)
            train_product += torch.dot(grad1, grad1)
            valid_product += torch.dot(grad2, grad2)

        cos_values.append((dot_product / torch.sqrt(train_product * valid_product)).cpu().item())

    sign_values = list(map(sign, cos_values))

    # Return the expectation as a Monte Carlo estimate from the samples in the list
    return sum(sign_values) / len(sign_values), sum(cos_values) / len(cos_values)


def gradient_variance(model, dataloader, criterion, device, gradient_pairs=1000):
    model_1, model_2 = _reparam(model), _reparam(model)
    limited_data_size = 2 * gradient_pairs
    limited_data_indices = random.sample(range(len(dataloader.dataset)), limited_data_size)
    limited_data = [dataloader.dataset[i] for i in limited_data_indices]
    limited_loader = DataLoader(limited_data, batch_size=2, shuffle=True, collate_fn=None)

    param_diff_squared = 0
    batches = 0
    for data, target in limited_loader:
        batches += 1

        pred1 = model_1(data.to(device, non_blocking=True))
        pred2 = model_2(data.to(device, non_blocking=True))

        loss1 = criterion(pred1, target.to(device))
        loss2 = criterion(pred2, target.to(device))

        model_1.zero_grad()
        model_2.zero_grad()
        loss1[0].backward()
        loss2[1].backward()

        params_1 = list(model_1.named_parameters())
        params_2 = list(model_2.named_parameters())
        params_1.sort(key=lambda element: element[0])
        params_2.sort(key=lambda element: element[0])

        for j in range(len(params_1)):
            grad1 = params_1[j][1].grad.flatten()
            grad2 = params_2[j][1].grad.flatten()
            param_diff_squared += torch.sum(torch.square(grad1 - grad2))
    return param_diff_squared / batches


if __name__ == "__main__":
    ce = nn.CrossEntropyLoss(reduction='none')
    batch_size = 10
    x1, y1 = torch.randn((batch_size, 3, 64, 64)), torch.randint(10, (batch_size,), dtype=torch.int64)
    x2, y2 = x1, torch.randint(10, (batch_size,), dtype=torch.int64)
    # x2, y2 = torch.randn((batch_size, 3, 64, 64)), torch.randint(10, (batch_size,), dtype=torch.int64)
    resnet = CustomResNet(depth=2, width=None, base_width=None, dataset_type=None)

    time0 = time.time()
    signs, coses = stiffness(resnet, x1, y1, x2, y2, ce)
    print("Processed {} points, cost {} seconds.".format(batch_size, time.time() - time0))
    print(signs)
    print(coses)
    print(sum(signs))
    print(sum(coses))
